@@ -27,22 +27,22 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 #include <panel-applet.h>
 #include <gdk/gdkkeysyms.h>
 
-#include "libindicator/indicator-object.h"
+#include <libido/libido.h>
+
+#include <libindicator/indicator-object.h>
+#include <libindicator/indicator-ng.h>
 #include "tomboykeybinder.h"
 
-static gchar * indicator_order[][2] = {
-  {"libappmenu.so", NULL},
-  {"libapplication.so", NULL},
-  {"libapplication.so", "gst-keyboard-xkb"},
-  {"libmessaging.so", NULL},
-  {"libpower.so", NULL},
-  {"libapplication.so", "bluetooth-manager"},
-  {"libnetwork.so", NULL},
-  {"libnetworkmenu.so", NULL},
-  {"libapplication.so", "nm-applet"},
-  {"libsoundmenu.so", NULL},
-  {"libdatetime.so", NULL},
-  {"libsession.so", NULL},
+static const gchar * indicator_order[][2] = {
+  {"libappmenu.so", NULL},                    /* indicator-appmenu" */
+  {"libapplication.so", NULL},                /* indicator-application" */
+  {"libprintersmenu.so", NULL},               /* indicator-printers */
+  {"libapplication.so", "gsd-keyboard-xkb"},  /* keyboard layout selector */
+  {"libmessaging.so", NULL},                  /* indicator-messages */
+  {"libbluetooth.so", NULL},                  /* indicator-bluetooth */
+  {"libnetwork.so", NULL},                    /* indicator-network */
+  {"libnetworkmenu.so", NULL},                /* indicator-network */
+  {"libapplication.so", "nm-applet"},         /* network manager */
   {NULL, NULL}
 };
 
@@ -57,6 +57,7 @@ static PanelAppletOrient orient;
 
 #define  IO_DATA_NAME                "indicator-name"
 #define  IO_DATA_ORDER_NUMBER        "indicator-order-number"
+#define  IO_DATA_MENUITEM_LOOKUP     "indicator-menuitem-lookup"
 
 static gboolean applet_fill_cb (PanelApplet * applet, const gchar * iid, gpointer data);
 
@@ -230,6 +231,7 @@ place_in_menu (GtkWidget *menubar,
   if (entry->name_hint != NULL) {
     const gchar *name = (const gchar *)g_object_get_data(G_OBJECT(io), IO_DATA_NAME);
     gint entry_position = name2order(name, entry->name_hint);
+    g_debug ("Placing %s (%s): %d", name, entry->name_hint, entry_position);
 
     /* If we don't find the entry, fall back to the indicator object's position */
     if (entry_position > -1)
@@ -362,17 +364,16 @@ accessible_desc_update (IndicatorObject * io, IndicatorObjectEntry * entry, GtkW
   return;
 }
 
-static void
-entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * menubar)
+static GtkWidget*
+create_menuitem (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * menubar)
 {
-  const char *indicator_name = (const gchar *)g_object_get_data(G_OBJECT(io), IO_DATA_NAME);
-  g_debug("Signal: Entry Added from %s", indicator_name);
-  gboolean something_visible = FALSE;
-  gboolean something_sensitive = FALSE;
+  GtkWidget * box;
+  GtkWidget * menuitem;
 
-  GtkWidget * menuitem = gtk_menu_item_new();
-  GtkWidget * box = (packdirection == GTK_PACK_DIRECTION_LTR) ?
-      gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 3) : gtk_box_new(GTK_ORIENTATION_VERTICAL, 3);
+  menuitem = gtk_menu_item_new();
+  box = (packdirection == GTK_PACK_DIRECTION_LTR)
+      ? gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 3)
+      : gtk_box_new (GTK_ORIENTATION_VERTICAL, 3);
 
   gtk_widget_add_events(GTK_WIDGET(menuitem), GDK_SCROLL_MASK);
 
@@ -389,18 +390,6 @@ entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * men
 
   if (entry->image != NULL) {
     gtk_box_pack_start(GTK_BOX(box), GTK_WIDGET(entry->image), FALSE, FALSE, 1);
-    if (gtk_widget_get_visible(GTK_WIDGET(entry->image))) {
-      something_visible = TRUE;
-    }
-
-    if (gtk_widget_get_sensitive(GTK_WIDGET(entry->image))) {
-      something_sensitive = TRUE;
-    }
-
-    g_signal_connect(G_OBJECT(entry->image), "show", G_CALLBACK(something_shown), menuitem);
-    g_signal_connect(G_OBJECT(entry->image), "hide", G_CALLBACK(something_hidden), menuitem);
-
-    g_signal_connect(G_OBJECT(entry->image), "notify::sensitive", G_CALLBACK(sensitive_cb), menuitem);
   }
   if (entry->label != NULL) {
     switch(packdirection) {
@@ -414,21 +403,10 @@ entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * men
         break;
       default:
         break;
-    }    
+    }
+    /* gtk_box_pack requires that the widget has no parent */
+    gtk_widget_unparent(GTK_WIDGET(entry->label));
     gtk_box_pack_start(GTK_BOX(box), GTK_WIDGET(entry->label), FALSE, FALSE, 1);
-
-    if (gtk_widget_get_visible(GTK_WIDGET(entry->label))) {
-      something_visible = TRUE;
-    }
-
-    if (gtk_widget_get_sensitive(GTK_WIDGET(entry->label))) {
-      something_sensitive = TRUE;
-    }
-
-    g_signal_connect(G_OBJECT(entry->label), "show", G_CALLBACK(something_shown), menuitem);
-    g_signal_connect(G_OBJECT(entry->label), "hide", G_CALLBACK(something_hidden), menuitem);
-
-    g_signal_connect(G_OBJECT(entry->label), "notify::sensitive", G_CALLBACK(sensitive_cb), menuitem);
   }
   gtk_container_add(GTK_CONTAINER(menuitem), box);
   gtk_widget_show(box);
@@ -439,6 +417,59 @@ entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * men
 
   place_in_menu(menubar, menuitem, io, entry);
 
+  return menuitem;
+}
+
+static void
+entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * menubar)
+{
+  const gchar * name;
+  GtkWidget * menuitem;
+  GHashTable * menuitem_lookup;
+  gboolean something_visible;
+  gboolean something_sensitive;
+
+  name = g_object_get_data (G_OBJECT(io), IO_DATA_NAME);
+  g_debug ("Signal: Entry Added from %s", name);
+
+  /* if the menuitem doesn't already exist, create it now */
+  menuitem_lookup = g_object_get_data (G_OBJECT(io), IO_DATA_MENUITEM_LOOKUP);
+  g_return_if_fail (menuitem_lookup != NULL);
+  menuitem = g_hash_table_lookup (menuitem_lookup, entry);
+  if (menuitem == NULL) {
+    menuitem = create_menuitem (io, entry, menubar);
+    g_hash_table_insert (menuitem_lookup, entry, menuitem);
+  }
+
+  /* connect the callbacks */
+  if (G_IS_OBJECT (entry->image)) {
+    g_object_connect (entry->image,
+                      "signal::show", G_CALLBACK(something_shown), menuitem,
+                      "signal::hide", G_CALLBACK(something_hidden), menuitem,
+                      "signal::notify::sensitive", G_CALLBACK(sensitive_cb), menuitem,
+                      NULL);
+  }
+  if (G_IS_OBJECT (entry->label)) {
+    g_object_connect (entry->label,
+                      "signal::show", G_CALLBACK(something_shown), menuitem,
+                      "signal::hide", G_CALLBACK(something_hidden), menuitem,
+                      "signal::notify::sensitive", G_CALLBACK(sensitive_cb), menuitem,
+                      NULL);
+  }
+
+  /* refresh based on visibility & sensitivity */
+  something_visible = FALSE;
+  something_sensitive = FALSE;
+  if (entry->image != NULL) {
+    GtkWidget * w = GTK_WIDGET (entry->image);
+    something_visible |= gtk_widget_get_visible (w);
+    something_sensitive |= gtk_widget_get_sensitive (w);
+  }
+  if (entry->label != NULL) {
+    GtkWidget * w = GTK_WIDGET (entry->label);
+    something_visible |= gtk_widget_get_visible (w);
+    something_sensitive |= gtk_widget_get_sensitive (w);
+  }
   if (something_visible) {
     if (entry->accessible_desc != NULL) {
       update_accessible_desc(entry, menuitem);
@@ -451,37 +482,37 @@ entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * men
 }
 
 static void
-entry_removed_cb (GtkWidget * widget, gpointer userdata)
-{
-  gpointer data = g_object_get_data(G_OBJECT(widget), MENU_DATA_INDICATOR_ENTRY);
-
-  if (data != userdata) {
-    return;
-  }
-
-  IndicatorObjectEntry * entry = (IndicatorObjectEntry *)data;
-  if (entry->label != NULL) {
-    g_signal_handlers_disconnect_by_func(G_OBJECT(entry->label), G_CALLBACK(something_shown), widget);
-    g_signal_handlers_disconnect_by_func(G_OBJECT(entry->label), G_CALLBACK(something_hidden), widget);
-    g_signal_handlers_disconnect_by_func(G_OBJECT(entry->label), G_CALLBACK(sensitive_cb), widget);
-  }
-  if (entry->image != NULL) {
-    g_signal_handlers_disconnect_by_func(G_OBJECT(entry->image), G_CALLBACK(something_shown), widget);
-    g_signal_handlers_disconnect_by_func(G_OBJECT(entry->image), G_CALLBACK(something_hidden), widget);
-    g_signal_handlers_disconnect_by_func(G_OBJECT(entry->image), G_CALLBACK(sensitive_cb), widget);
-  }
-
-  gtk_widget_destroy(widget);
-  return;
-}
-
-static void
-entry_removed (IndicatorObject * io G_GNUC_UNUSED, IndicatorObjectEntry * entry,
+entry_removed (IndicatorObject * io,
+               IndicatorObjectEntry * entry,
                gpointer user_data)
 {
+  GtkWidget * menuitem;
+  GHashTable * menuitem_lookup;
+
   g_debug("Signal: Entry Removed");
 
-  gtk_container_foreach(GTK_CONTAINER(user_data), entry_removed_cb, entry);
+  menuitem_lookup = g_object_get_data (G_OBJECT(io), IO_DATA_MENUITEM_LOOKUP);
+  g_return_if_fail (menuitem_lookup != NULL);
+  menuitem = g_hash_table_lookup (menuitem_lookup, entry);
+  g_return_if_fail (menuitem != NULL);
+
+  /* disconnect the callbacks */
+  if (G_IS_OBJECT (entry->label)) {
+    g_object_disconnect (entry->label,
+                         "any-signal", G_CALLBACK(something_shown), menuitem,
+                         "any-signal", G_CALLBACK(something_hidden), menuitem,
+                         "any-signal", G_CALLBACK(sensitive_cb), menuitem,
+                         NULL);
+  }
+  if (G_IS_OBJECT (entry->image)) {
+    g_object_disconnect (entry->image,
+                         "any-signal", G_CALLBACK(something_shown), menuitem,
+                         "any-signal", G_CALLBACK(something_hidden), menuitem,
+                         "any-signal", G_CALLBACK(sensitive_cb), menuitem,
+                         NULL);
+  }
+
+  gtk_widget_hide (menuitem);
 
   return;
 }
@@ -575,10 +606,47 @@ update_accessible_desc(IndicatorObjectEntry * entry, GtkWidget * menuitem)
   return;
 }
 
+static void load_indicator(GtkWidget * menubar, IndicatorObject *object, const gchar *name) {
+	GObject * o;
+	GList *entries, *entry;
+
+	/* Set the environment it's in */
+	indicator_object_set_environment(object, (GStrv)indicator_env);
+
+	/* Attach the 'name' to the object */
+	o = G_OBJECT (object);
+	g_object_set_data_full(o, IO_DATA_MENUITEM_LOOKUP, g_hash_table_new (g_direct_hash, g_direct_equal), (GDestroyNotify)g_hash_table_destroy);
+	g_object_set_data_full(o, IO_DATA_NAME, g_strdup(name), g_free);
+	
+	int pos = 5000 - indicator_object_get_position(object);
+	if (pos > 5000) {
+	    pos = name2order(name, NULL);
+	}
+	
+	g_object_set_data(o, IO_DATA_ORDER_NUMBER, GINT_TO_POINTER(pos));
+
+	/* Connect to its signals */
+	g_signal_connect(o, INDICATOR_OBJECT_SIGNAL_ENTRY_ADDED,   G_CALLBACK(entry_added),    menubar);
+	g_signal_connect(o, INDICATOR_OBJECT_SIGNAL_ENTRY_REMOVED, G_CALLBACK(entry_removed),  menubar);
+	g_signal_connect(o, INDICATOR_OBJECT_SIGNAL_ENTRY_MOVED,   G_CALLBACK(entry_moved),    menubar);
+	g_signal_connect(o, INDICATOR_OBJECT_SIGNAL_MENU_SHOW,     G_CALLBACK(menu_show),      menubar);
+	g_signal_connect(o, INDICATOR_OBJECT_SIGNAL_ACCESSIBLE_DESC_UPDATE, G_CALLBACK(accessible_desc_update), menubar);
+
+	/* Work on the entries */
+	entries = indicator_object_get_entries(object);
+
+	for (entry = entries; entry != NULL; entry = g_list_next(entry)) {
+		IndicatorObjectEntry * entrydata = (IndicatorObjectEntry *) entry->data;
+		entry_added(object, entrydata, menubar);
+	}
+
+	g_list_free(entries);
+}
 
 static gboolean
 load_module (const gchar * name, GtkWidget * menubar)
 {
+  
   g_debug("Looking at Module: %s", name);
   g_return_val_if_fail(name != NULL, FALSE);
 
@@ -593,34 +661,126 @@ load_module (const gchar * name, GtkWidget * menubar)
   IndicatorObject * io = indicator_object_new_from_file(fullpath);
   g_free(fullpath);
 
-  /* Set the environment it's in */
-  indicator_object_set_environment(io, (GStrv)indicator_env);
-
-  /* Attach the 'name' to the object */
-  g_object_set_data_full(G_OBJECT(io), IO_DATA_NAME, g_strdup(name), g_free);
-  g_object_set_data(G_OBJECT(io), IO_DATA_ORDER_NUMBER, GINT_TO_POINTER(name2order(name, NULL)));
-
-  /* Connect to its signals */
-  g_signal_connect(G_OBJECT(io), INDICATOR_OBJECT_SIGNAL_ENTRY_ADDED,   G_CALLBACK(entry_added),    menubar);
-  g_signal_connect(G_OBJECT(io), INDICATOR_OBJECT_SIGNAL_ENTRY_REMOVED, G_CALLBACK(entry_removed),  menubar);
-  g_signal_connect(G_OBJECT(io), INDICATOR_OBJECT_SIGNAL_ENTRY_MOVED,   G_CALLBACK(entry_moved),    menubar);
-  g_signal_connect(G_OBJECT(io), INDICATOR_OBJECT_SIGNAL_MENU_SHOW,     G_CALLBACK(menu_show),      menubar);
-  g_signal_connect(G_OBJECT(io), INDICATOR_OBJECT_SIGNAL_ACCESSIBLE_DESC_UPDATE, G_CALLBACK(accessible_desc_update), menubar);
-
-  /* Work on the entries */
-  GList * entries = indicator_object_get_entries(io);
-  GList * entry = NULL;
-
-  for (entry = entries; entry != NULL; entry = g_list_next(entry)) {
-    IndicatorObjectEntry * entrydata = (IndicatorObjectEntry *)entry->data;
-    entry_added(io, entrydata, menubar);
-  }
-
-  g_list_free(entries);
+  load_indicator(menubar, io, name);
 
   return TRUE;
 }
 
+static void load_modules(GtkWidget *menubar, gint *indicators_loaded) {
+	if  (g_file_test(INDICATOR_DIR, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))) {
+		GDir * dir = g_dir_open(INDICATOR_DIR, 0, NULL);
+
+		const gchar * name;
+		gint count = 0;
+		while ((name = g_dir_read_name(dir)) != NULL) {
+			
+			#ifdef INDICATOR_APPLET_APPMENU
+				if (g_strcmp0(name, "libappmenu.so")) {
+					continue;
+				}
+			#else
+				if (!g_strcmp0(name, "libappmenu.so")) {
+					continue;
+				}
+			#endif
+			
+			#ifdef INDICATOR_APPLET
+				if (!g_strcmp0(name, "libsession.so")) {
+					continue;
+				}
+				if (!g_strcmp0(name, "libme.so")) {
+					continue;
+				}
+				if (!g_strcmp0(name, "libdatetime.so")) {
+					continue;
+				}
+			#endif
+			
+			#ifdef INDICATOR_APPLET_SESSION
+				if (g_strcmp0(name, "libsession.so") && g_strcmp0(name, "libme.so")) {
+					continue;
+				}
+			#endif
+
+			if (load_module(name, menubar)) {
+				count++;
+			}
+		}
+
+		*indicators_loaded += count;
+		
+		g_dir_close (dir);
+	}
+}
+
+#define INDICATOR_SERVICE_DIR "/usr/share/unity/indicators"
+
+static void load_indicators_from_indicator_files(GtkWidget *menubar, gint *indicators_loaded) {
+	GDir *dir;
+	const gchar *name;
+	GError *error = NULL;
+
+	dir = g_dir_open (INDICATOR_SERVICE_DIR, 0, &error);
+
+	if (!dir) {
+		g_warning ("unable to open indicator service file directory: %s", error->message);
+  		g_error_free (error);
+		
+  		return;
+	}
+	
+	gint count = 0;
+	while ((name = g_dir_read_name (dir))) {
+		gchar *filename;
+		IndicatorNg *indicator;
+
+		filename = g_build_filename (INDICATOR_SERVICE_DIR, name, NULL);
+		indicator = indicator_ng_new_for_profile (filename, "desktop", &error);
+		g_free (filename);
+
+		#ifdef INDICATOR_APPLET_APPMENU
+			if (g_strcmp0(name, "com.canonical.indicator.appmenu")) {
+				continue;
+			}
+		#else
+			if (!g_strcmp0(name, "com.canonical.indicator.appmenu")) {
+				continue;
+			}
+		#endif
+		
+		#ifdef INDICATOR_APPLET
+			if (!g_strcmp0(name, "com.canonical.indicator.session")) {
+				continue;
+			}
+			if (!g_strcmp0(name, "com.canonical.indicator.me")) {
+				continue;
+			}
+			if (!g_strcmp0(name, "com.canonical.indicator.datetime")) {
+				continue;
+			}
+		#endif
+		
+		#ifdef INDICATOR_APPLET_SESSION
+			if (g_strcmp0(name, "com.canonical.indicator.session") && g_strcmp0(name, "com.canonical.indicator.me")) {
+				continue;
+			}
+		#endif
+
+		if (indicator) {
+			g_debug ("loading indicator: %s", name);
+			load_indicator(menubar,INDICATOR_OBJECT (indicator), name);
+			count++;
+		}else{
+			g_warning ("unable to load '%s': %s", name, error->message);
+			g_clear_error (&error);
+		}
+	}
+
+	*indicators_loaded += count;
+
+	g_dir_close (dir);
+}
+
 static void
 hotkey_filter (char * keystring, gpointer data)
 {
@@ -832,6 +992,8 @@ static gboolean
 applet_fill_cb (PanelApplet * applet, const gchar * iid G_GNUC_UNUSED,
                 gpointer data G_GNUC_UNUSED)
 {
+  ido_init();
+  
   static const GtkActionEntry menu_actions[] = {
     {"About", GTK_STOCK_ABOUT, N_("_About"), NULL, NULL, G_CALLBACK(about_cb)}
   };
@@ -923,43 +1085,9 @@ applet_fill_cb (PanelApplet * applet, const gchar * iid G_GNUC_UNUSED,
   /* Add in filter func */
   tomboy_keybinder_bind(hotkey_keycode, hotkey_filter, menubar);
 
-  /* load 'em */
-  if (g_file_test(INDICATOR_DIR, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))) {
-    GDir * dir = g_dir_open(INDICATOR_DIR, 0, NULL);
-
-    const gchar * name;
-    while ((name = g_dir_read_name(dir)) != NULL) {
-#ifdef INDICATOR_APPLET_APPMENU
-      if (g_strcmp0(name, "libappmenu.so")) {
-        continue;
-      }
-#else
-      if (!g_strcmp0(name, "libappmenu.so")) {
-        continue;
-      }
-#endif
-#ifdef INDICATOR_APPLET
-      if (!g_strcmp0(name, "libsession.so")) {
-        continue;
-      }
-      if (!g_strcmp0(name, "libme.so")) {
-        continue;
-      }
-      if (!g_strcmp0(name, "libdatetime.so")) {
-        continue;
-      }
-#endif
-#ifdef INDICATOR_APPLET_SESSION
-      if (g_strcmp0(name, "libsession.so") && g_strcmp0(name, "libme.so")) {
-        continue;
-      }
-#endif
-      if (load_module(name, menubar)) {
-        indicators_loaded++;
-      }
-    }
-    g_dir_close (dir);
-  }
+	/* load indicators */
+	load_modules(menubar, &indicators_loaded);
+	load_indicators_from_indicator_files(menubar, &indicators_loaded);
 
   if (indicators_loaded == 0) {
     /* A label to allow for click through */
@@ -976,4 +1104,3 @@ applet_fill_cb (PanelApplet * applet, const gchar * iid G_GNUC_UNUSED,
 
   return TRUE;
 }
-
